# Copyright (c) 2025, UChicago Argonne, LLC
# BSD OPEN SOURCE LICENSE. Full license can be found in LICENSE.md
from sqlite3 import Connection

import shapely.wkb
from shapely.geometry import MultiLineString

from polaris.network.constants import constants, ROUTE_ID_MULTIPLIER, AGENCY_MULTIPLIER
from polaris.network.transit.transit_elements.basic_element import BasicPTElement


class Route(BasicPTElement):
    """Transit route element to feed into Transit_routes

    :Database class members:

    * route_id (:obj:`str`): ID of this route, starting with the agency prefix ID
    * route_short_name (:obj:`str`): Short name as found in the GTFS feed
    * route_long_name (:obj:`str`): Long name as found in the GTFS feed
    * route_desc (:obj:`str`): Route description as found in the GTFS feed
    * route_type (:obj:`int`):  Route type (mode) for this transit link
    * route_url (:obj:`str`): Route URL as found in the GTFS feed
    * route_color (:obj:`str`): Route color for mapping as found in the GTFS feed
    * route_text_color (:obj:`str`): Route color (text) for mapping as found in the GTFS feed
    * route_sort_order (:obj:`int`): Route rendering order as found in the GTFS feed
    * agency_id (:obj:`str`): Agency ID
    * seated_capacity (:obj:`float`): Vehicle seated capacity for this route
    * design_capacity (:obj:`float`): Vehicle design capacity for this route
    * total_capacity (:obj:`float`): Total vehicle capacity for this route"""

    def __init__(self, agency_id):
        self.route_id = -1
        self.route = ""
        self.route_short_name = ""
        self.route_long_name = ""
        self.route_desc = ""
        self.route_type = 0
        self.route_url = ""
        self.route_color = ""
        self.route_text_color = ""
        self.route_sort_order = 0
        self.agency_id = agency_id

        # Not part of GTFS
        self.seated_capacity = 0
        self.design_capacity = 0
        self.total_capacity = 0
        self.shape = MultiLineString()
        self.srid = -1
        self.number_of_cars = 0
        self.__sql = """insert into Transit_Routes (route_id, route, agency_id, shortname, longname, description,
                                                    "type", seated_capacity, design_capacity, total_capacity,
                                                     number_of_cars{}) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?{});"""
        self.sql = self.__sql

    def populate(self, record: tuple, headers: list) -> None:
        for key, value in zip(headers, record):
            if key not in self.__dict__.keys():
                raise KeyError(f"{key} field in Routes.txt is unknown field for that file on GTFS")

            # We convert route_id into route, as the the GTFS header for it is not maintained in our database
            key = "route" if key == "route_id" else key
            self.__dict__[key] = value

    def save_to_database(self, conn: Connection, commit=True) -> None:
        """Saves route to the database"""

        data = self.data
        conn.execute(self.sql, data)
        if commit:
            conn.commit()

    @property
    def data(self):
        data = [
            self.route_id,
            self.route,
            self.agency_id,
            self.route_short_name,
            self.route_long_name,
            self.route_desc,
            int(self.route_type),
            self.seated_capacity,
            self.design_capacity,
            self.total_capacity,
            self.number_of_cars,
        ]
        if self.shape is None:
            shape = ""
            geo_fld = ""
        else:
            geo_fld = ",geo "
            shape = ", ST_Multi(GeomFromWKB(?, ?))"
            data.extend([self.shape.wkb, self.srid])
        self.sql = self.__sql.format(geo_fld, shape)
        return data

    def from_database(self, conn: Connection, route_id: int):
        sql = """select route, agency_id, shortname, longname, description, "type", seated_capacity,
                 design_capacity, total_capacity, number_of_cars, ST_AsBinary(geo) from Transit_Routes
                 where route_id=?"""
        data = conn.execute(sql, [route_id]).fetchone()
        (
            self.route,
            self.agency_id,
            self.route_short_name,
            self.route_long_name,
            self.route_desc,
            self.route_type,
            self.seated_capacity,
            self.design_capacity,
            self.total_capacity,
            self.number_of_cars,
            self.shape,
        ) = data
        self.shape = shapely.wkb.loads(self.shape)
        self.route_id = route_id
        return self

    def get_route_id(self):
        c = constants()
        val = c.routes.get(self.agency_id, AGENCY_MULTIPLIER * self.agency_id)
        c.routes[self.agency_id] = val + ROUTE_ID_MULTIPLIER
        self.route_id = c.routes[self.agency_id]
