# Copyright (c) 2025, UChicago Argonne, LLC
# BSD OPEN SOURCE LICENSE. Full license can be found in LICENSE.md
from sqlite3 import Connection
from typing import Optional

from shapely.geometry import LineString

from polaris.network.constants import constants, TRIP_ID_MULTIPLIER
from polaris.network.starts_logging import logger
from polaris.network.transit.transit_elements.basic_element import BasicPTElement


class Trip(BasicPTElement):
    """Transit trips read from trips.txt

    :GTFS class members:

    * trip (:obj:`str`): Trip ID as read from the GTFS feed
    * route (:obj:`str`): Route ID as read from the GTFS feed
    * service_id (:obj:`str`): Service ID as read from the GTFS feed
    * trip_headsign (:obj:`str`): Trip headsign as read from the GTFS feed
    * trip_short_name (:obj:`str`): Trip short name as read from the GTFS feed
    * direction_id (:obj:`int`): Direction ID as read from the GTFS feed
    * block_id (:obj:`int`): Block ID as read from the GTFS feed
    * bikes_allowed (:obj:`int`): Bikes allowed flag as read from the GTFS feed
    * wheelchair_accessible (:obj:`int`): Wheelchair accessibility flag as read from the GTFS feed
    * shape_id (:obj:`str`): Shape ID as read from the GTFS feed

    :Processing class members:
    * trip_id (:obj:`int`): Unique trip_id as it will go into the database
    * route_id (:obj:`int`): Unique Route ID as will be available in the routes table
    * pattern_id (:obj:`int`): Unique Pattern ID for this route/stop-pattern as it will go into the database
    * pattern_hash (:obj:`str`): Pattern ID derived from stops for this route/stop-pattern
    * arrivals (:obj:`List[int]`): Sequence of arrival at stops for this trip
    * departures (:obj:`List[int]`): Sequence of departures from stops for this trip
    * stops (:obj:`List[Stop]`): Sequence of stops for this trip
    * shape (:obj:`LineString`): Shape for this trip. Directly from shapes.txt or rebuilt from sequence of stops
    """

    def __init__(self):
        self.route_id = ""
        self.service_id = ""
        self.trip = ""
        self.trip_id = -1
        self.trip_headsign = ""
        self.trip_short_name = ""
        self.block_id = ""
        self.shape_id = ""
        self.direction_id = 0
        self.wheelchair_accessible = 0
        self.bikes_allowed = 0

        # Not from GTFS
        self.pattern_id = 0
        self.pattern_hash = ""
        self.arrivals = []
        self.departures = []
        self.stops = []
        self.shape = LineString()
        self._stop_based_shape = LineString()
        self.seated_capacity: Optional[int] = None
        self.design_capacity: Optional[int] = None
        self.total_capacity: Optional[int] = None
        self.is_artic = 0
        self.source_time = []

    def _populate(self, record: tuple, headers: list) -> None:
        for key, value in zip(headers, record):
            if key not in self.__dict__.keys():
                raise KeyError(f"{key} field in Trips.txt is unknown field for that file on GTFS")
            key = "trip" if key == "trip_id" else key
            key = "route" if key == "route_id" else key
            self.__dict__[key] = value

    def save_to_database(self, conn: Connection, commit=True) -> None:
        """Saves Transit trip to the database"""
        logger.debug(f"Saving {self.trip_id}/{self.trip} for pattern {self.pattern_id}")
        sql = """insert into Transit_Trips (trip_id, trip,"dir", pattern_id, seated_capacity, design_capacity,
                                             total_capacity, is_artic) values (?, ?, ?, ?, ?, ?, ?, ?);"""
        data = [
            self.trip_id,
            self.trip,
            int(self.direction_id),
            self.pattern_id,
            self.seated_capacity,
            self.design_capacity,
            self.total_capacity,
            self.is_artic,
        ]
        conn.execute(sql, data)

        sql = """insert into Transit_Trips_Schedule(trip_id, "index", arrival, departure, time_source)
                                            values (?, ?, ?, ?, ?)"""
        data = []
        for i, (arr, dep, st) in enumerate(zip(self.arrivals, self.departures, self.source_time)):
            data.append([self.trip_id, i, arr, dep, st])
        conn.executemany(sql, data)
        if commit:
            conn.commit()

    def get_trip_id(self):
        c = constants()
        self.trip_id = c.trips.get(self.pattern_id, self.pattern_id) + TRIP_ID_MULTIPLIER
        c.trips[self.pattern_id] = self.trip_id

    def __lt__(self, other):
        return self.departures[0] < other.departures[0]
