# Copyright (c) 2025, UChicago Argonne, LLC
# BSD OPEN SOURCE LICENSE. Full license can be found in LICENSE.md
import geopandas as gpd

from polaris.utils.database.data_table_access import DataTableAccess
from polaris.prepare.supply_tables.ev_chargers.ev_charging import ev_chargers
from polaris.prepare.supply_tables.network.net_constructor import NetworkConstructor
from polaris.prepare.supply_tables.utils.state_county_candidates import get_state_counties
from polaris.prepare.supply_tables.zones.zoning import add_zoning


class Populator:
    def __init__(self, supply_path) -> None:
        self.supply_path = supply_path
        self._state_counties = gpd.GeoDataFrame()

    def add_zoning_system(self, model_area: gpd.GeoDataFrame, zone_level: str, census_api_key: str, year: int = 2021):
        """Create zoning system based on census subdivisions. Population data is extracted from The ACS 5 year. For
        mode details see https://github.com/datamade/census

        Args:
            model_area (GeoDataFrame): GeoDataFrame containing polygons with the model area
            zone_level (str): Census subdivision level to use for zoning -> tracts or block_groups
            year (int): Year of the population data. Defaults to 2021
            census_api_key (str): API key for the census.
        """

        assert zone_level in ["tracts", "block_groups"]

        if census_api_key is None:
            raise ValueError("No Census API provided. Please obtain one at https://api.census.gov/data/key_signup.html")

        self._state_counties = get_state_counties(model_area)
        add_zoning(model_area, self.state_counties, zone_level, self.supply_path, year, census_api_key)

    def add_ev_chargers(self, nrel_api_key: str, max_dist=100, clustering_attempts=5):
        """Adds EV Chargers

        Args:
            nrel_api_key (str): API key for the NREL API to retrieve location of EV chargers.
            max_dist (float):
            clustering_attempts (int):
        """
        if nrel_api_key is None:
            raise ValueError("No Census API provided. Please obtain one at https://developer.nrel.gov/signup/")

        assert max_dist > 0, "Minimum clustering must be non-negative. Use 0.1 if you must. That's in meters"

        model_area = DataTableAccess(self.supply_path).get("Zone")
        ev_chargers(model_area, self.supply_path, nrel_api_key, max_dist, clustering_attempts)

    def add_parking(
        self,
        facility_types: tuple = ("surface", "underground", "multi-storey", "rooftop"),
        sample_rate: float = 1.0,
    ):
        """Adds Parking facilities from data coming from OpenStreetMap

        Args:
           facility_types (str): Facility type according to OSM
        """
        from polaris.prepare.supply_tables.park.parking import add_parking

        add_parking(self.supply_path, facility_types, sample_rate)

    def add_locations(
        self, control_totals: gpd.GeoDataFrame, census_api_key: str, residential_sample_rate=0.05, other_sample_rate=1.0
    ):
        from polaris.prepare.supply_tables.locations.locations import add_locations

        add_locations(
            supply_path=self.supply_path,
            state_counties=self.state_counties,
            control_totals=control_totals,
            census_api_key=census_api_key,
            residential_sample_rate=residential_sample_rate,
            other_sample_rate=other_sample_rate,
        )

    def create_network(
        self,
        simplification_parameters={  # noqa: B006
            "simplify": True,
            "keep_transit_links": True,
            "accessibility_level": "zone",
            "maximum_network_capacity": False,
        },
        imputation_parameters={"algorithm": "knn", "max_iter": 10},  # noqa: B006
    ) -> NetworkConstructor:
        """ """
        if self._state_counties.empty:
            self._state_counties = get_state_counties(model_area=DataTableAccess(self.supply_path).get("Zone"))

        return NetworkConstructor(
            polaris_network_path=self.supply_path,
            simplification_parameters=simplification_parameters,
            imputation_parameters=imputation_parameters,
            state_counties=self._state_counties,
        )

    @property
    def state_counties(self):
        if self._state_counties.empty:
            model_area = DataTableAccess(self.supply_path).get("Zone")
            self._state_counties = get_state_counties(model_area)
        return self._state_counties
